"""Contains the ForeignModelWrapper class."""
from . import ModelBase
from bidon.util.transform import get_obj


class ForeignModelWrapper(ModelBase):
  timestamps = None
  table_name = None
  primary_key_name = None
  primary_key_is_auto = False
  transform_args = None

  @classmethod
  def create(cls, source, *, transform_args=None):
    if transform_args is None:
      transform_args = cls.transform_args

    return cls(get_obj(source, *transform_args))

  @classmethod
  def map(cls, sources, *, transform_args=None):
    for idx, source in enumerate(sources):
      try:
        yield cls.create(source, transform_args=transform_args)
      except Exception as ex:
        raise Exception("An error occurred with item {0}".format(idx)) from ex




import unittest

from bidon.data import SQLWriter


class DataSQLWriterTestCase(unittest.TestCase):
  def setUp(self):
    self.sqlw = SQLWriter("%({0})s", "%s", "{0}::{1}", sort_columns=True)

  def _normalize(self, s, joiner):
    return set(s.split(joiner))

  def test_to_placeholder(self):
    self.assertEqual(self.sqlw.to_placeholder("name"), "%(name)s")
    self.assertEqual(self.sqlw.to_placeholder(None), "%s")
    self.assertEqual(self.sqlw.to_placeholder("name", "int"), "%(name)s::int")
    self.assertEqual(self.sqlw.to_placeholder(None, "citext"), "%s::citext")

  def test_to_tuple(self):
    self.assertEqual(self.sqlw.to_tuple(["a","b","c"]), "(a, b, c)")

  def test_value_comparisons(self):
    values = dict(first_name="Bob", last_name="Smith")
    comps = self.sqlw.value_comparisons(values)
    target = ["first_name = %(first_name)s", "last_name = %(last_name)s"]
    self.assertEqual(set(comps), set(target))

  def test_parse_constraints(self):
    sql, params = self.sqlw.parse_constraints("last_name = 'Smith'")
    self.assertEqual(sql, "last_name = 'Smith'")
    self.assertIsNone(params)
    sql, params = self.sqlw.parse_constraints(dict(last_name="Smith"))
    self.assertEqual(sql, "last_name = %(last_name)s")
    self.assertEqual(params, dict(last_name="Smith"))
    sql, params = self.sqlw.parse_constraints(("last_name = %s and first_name = %s", ["Smith", "John"]))
    self.assertEqual(sql, "last_name = %s and first_name = %s")
    self.assertEqual(params, ["Smith", "John"])
    self.sqlw.sort_columns = False
    values = (("id", 10, " <= "), ("last_name", "Smith"), ("age", (40, 50), "between"), ("first_name", ("John", "Susie"), "in"))
    sql, params = self.sqlw.parse_constraints(values)
    self.assertEqual(sql, "id <= %s and last_name = %s and age between %s and %s and first_name in (%s, %s)")
    self.sqlw.sort_columns = True
    sql, params = self.sqlw.parse_constraints(values)
    self.assertEqual(sql, "age between %s and %s and first_name in (%s, %s) and id <= %s and last_name = %s")
    self.assertEqual(params, [10, "Smith", 40, 50, "John", "Susie"])
    self.assertEqual("last_name is %s", self.sqlw.parse_constraints([("last_name", None)])[0])
    self.assertEqual("last_name is not %s", self.sqlw.parse_constraints([("last_name", None, "!=")])[0])
    self.assertEqual("last_name is not %s", self.sqlw.parse_constraints([("last_name", None, "is not")])[0])
    self.assertEqual("last_name is %(last_name)s", self.sqlw.parse_constraints(dict(last_name=None))[0])
    self.assertEqual("not (schedule is %s)", self.sqlw.parse_constraints([("schedule", None, "not(is)")])[0])
    self.assertEqual("not (schedule is not %s)", self.sqlw.parse_constraints([("schedule", None, "not(is not)")])[0])

  def test_get_find_all_query(self):
    vals = dict(first_name="Trey")
    self.assertEqual(self.sqlw.get_find_all_query("people")[0], "select * from people where 1 = 1")
    self.assertEqual(self.sqlw.get_find_all_query("people", vals)[0], "select * from people where first_name = %(first_name)s")
    self.assertEqual(self.sqlw.get_find_all_query("people", vals, columns="id, first_name")[0], "select id, first_name from people where first_name = %(first_name)s")
    self.assertEqual(self.sqlw.get_find_all_query("people", vals, columns=["id", "first_name"], order_by="last_name desc")[0], "select id, first_name from people where first_name = %(first_name)s order by last_name desc")



"""Contains a freezable configuration class."""

class Configuration(object):
  """A freezable configuration object. You can set it up, and update it a few times, and then
  freeze it to prevent new values from being added, or existing values from being edited.
  """
  def __init__(self, **kwargs):
    """Initializes the Configuration class. Each parameter becomes a property on the instance."""
    self._frozen_ = False
    self.update(**kwargs)

  @property
  def is_frozen(self):
    return hasattr(self, "_frozen_") and self._frozen_

  def update(self, **kwargs):
    """Creates or updates a property for the instance for each parameter."""
    for key, value in kwargs.items():
      setattr(self, key, value)

  def freeze(self):
    """Freezes the instance, preventing new attributes from being added or edited."""
    if not self.is_frozen:
      self._frozen_ = True

  def __setattr__(self, k, v):
    """Overrides setattr to enforze the frozen attribute."""
    if self.is_frozen:
      raise Exception("Configuration is frozen")
    super().__setattr__(k, v)



import unittest

from bidon.field_mapping import FieldMapping


class FMSource(object):
  def __init__(self, name, age, house):
    self.name = name
    self.age = age
    self.house = house


class FMDest(object):
  def __init__(self, first_name, last_name, age, house_name):
    self.first_name = first_name
    self.last_name = last_name
    self.age = age
    self.house_name = house_name


fn_fm = FieldMapping("name", "first_name", lambda v: v.split(", ")[1])
ln_fm = FieldMapping("name", "last_name", lambda v: v.split(", ")[0])
age_fm = FieldMapping("age")
house_fm = FieldMapping("house", "house_name", lambda v: "{0} House".format(v), False)


class FieldMappingTestCase(unittest.TestCase):
  def test_get_value(self):
    s1 = FMSource("Ravenclaw, Rowena", 40, "Ravenclaw")
    self.assertEqual(fn_fm.get_value(s1), "Rowena")
    self.assertEqual(ln_fm.get_value(s1), "Ravenclaw")
    self.assertEqual(age_fm.get_value(s1), 40)
    self.assertEqual(house_fm.get_value(s1), "Ravenclaw House")
    s2 = { "age": 55 }
    self.assertEqual(age_fm.get_value(s2), 55)
    s3 = { }
    self.assertRaises(Exception, lambda: age_fm.get_value(s3))
    self.assertIsNone(house_fm.get_value(s3))

  def test_transfer(self):
    s1 = FMSource("Gryffindor, Godric", 40, "Gryffindor")
    s2 = dict(name="Gryffindor, Godric", age=40)
    fms = [fn_fm, ln_fm, age_fm, house_fm]

    d1 = FieldMapping.transfer(fms, s1, lambda d: FMDest(**d))
    self.assertEqual(d1.first_name, "Godric")
    self.assertEqual(d1.last_name, "Gryffindor")
    self.assertEqual(d1.age, 40)
    self.assertEqual(d1.house_name, "Gryffindor House")

    d2 = FieldMapping.transfer(fms, s1, dict)
    self.assertEqual(d2, dict(first_name="Godric", last_name="Gryffindor", age=40, house_name="Gryffindor House"))

    d3 = FieldMapping.transfer(fms, s2, lambda x: x)
    self.assertEqual(d3, dict(first_name="Godric", last_name="Gryffindor", age=40, house_name=None))

  def test_transfer_all(self):
    sources = [
      FMSource("McGonagall, Minerva", 40, "Gryffindor"),
      FMSource("Flitwick, Filius", 40, "Ravenclaw"),
      FMSource("Sprout, Pomona", 40, "Hufflepuff"),
      FMSource("Snape, Severus", 40, "Slytherin")
    ]
    fms = [fn_fm, ln_fm, age_fm, house_fm]

    d1 = list(FieldMapping.transfer_all(fms, sources, lambda d: FMDest(**d)))
    self.assertEqual(len(d1), 4)
    self.assertIsInstance(d1[0], FMDest)
    self.assertEqual(d1[0].first_name, "Minerva")

    d2 = list(FieldMapping.transfer_all(fms, sources))
    self.assertEqual(len(d2), 4)
    self.assertIsInstance(d2[0], dict)
    self.assertEqual(d2[-1]["last_name"], "Snape")

    d3 = list(FieldMapping.transfer_all(fms, sources, FieldMapping.get_namedtuple_factory(fms)))
    self.assertEqual(len(d3), 4)
    self.assertEqual(d3[0].first_name, "Minerva")

  def test_get_namedtuple(self):
    fms = [fn_fm, ln_fm, age_fm, house_fm]
    nt = FieldMapping.get_namedtuple(fms)
    for fm in fms:
      self.assertTrue(hasattr(nt, fm.destination_name))

  def test_get_namedtuple_factory(self):
    fms = [fn_fm, ln_fm, age_fm, house_fm]
    ntf = FieldMapping.get_namedtuple_factory(fms)
    nt = ntf(dict(first_name="Vernon", last_name="Dursley", age=35, house_name="Privet Drive"))
    for fm in fms:
      self.assertTrue(hasattr(nt, fm.destination_name))



import unittest

from bidon.configuration import Configuration

class ConfigurationTestCase(unittest.TestCase):
  def test_configuration(self):
    c = Configuration(name="Trey", age=32, kid_count=2)
    c.update(name="Trey Cucco", spouse="Julie")
    self.assertFalse(c.is_frozen)
    c.freeze()
    self.assertEqual(c.name, "Trey Cucco")
    self.assertEqual(c.age, 32)
    self.assertEqual(c.kid_count, 2)
    self.assertEqual(c.spouse, "Julie")

    with self.assertRaises(Exception):
      c.name = "Trey"
    with self.assertRaises(Exception):
      c.last_name = "Cucco"
    with self.assertRaises(Exception):
      c.update(siblings=1)




import re
import unittest
from collections import namedtuple

from bidon.data import ModelBase, Validation, Validator, ModelAccess

from tests import get_model_access


class Person(ModelBase):
  table_name = "people"
  attrs = {
    "first_name": None,
    "last_name": None,
    "age": 0}
  validator = Validator([
    Validation.is_unique("first_name", scope=["last_name"])])


class DataValidationTestCase(unittest.TestCase):
  def setUp(self):
    self.model = Person(first_name="Lord", last_name="Voldemort", age=68)

  def test_is_present(self):
    v = Validation.is_present("first_name")
    self.assertTrue(v.is_valid(self.model)[0])
    self.model.first_name = None
    self.assertFalse(v.is_valid(self.model)[0])

  def test_is_length(self):
    v = Validation.is_length("first_name", min_length=2)
    self.assertTrue(v.is_valid(self.model)[0])
    self.model.first_name = "L"
    self.assertFalse(v.is_valid(self.model)[0])
    self.model.first_name = None
    self.assertFalse(v.is_valid(self.model)[0])
    v = Validation.is_length("first_name", present_optional=True)
    self.assertTrue(v.is_valid(self.model)[0])
    v = Validation.is_length("last_name", max_length=5)
    self.assertFalse(v.is_valid(self.model)[0])

  def test_matches(self):
    v = Validation.matches("last_name", re.compile(r"^[A-Z]"))
    self.assertTrue(v.is_valid(self.model)[0])
    self.model.last_name = self.model.last_name.lower()
    self.assertFalse(v.is_valid(self.model)[0])

  def test_is_unique(self):
    with get_model_access() as ma:
      self.assertTrue(self.model.validate(ma))
      ma.insert_model(self.model)
      self.assertTrue(self.model.validate(ma))
      m2 = Person(first_name="Lord", last_name="Voldemort")
      self.assertFalse(m2.validate(ma))

      class P2(ModelBase):
        table_name = "people"
        attrs = dict(first_name=None, last_name=None, age=0)
        validator = Validator([
          Validation.is_unique("first_name", comparison_operators=[" = "])
        ])

      p2 = P2(first_name="Lord", last_name="Churchhill")
      self.assertFalse(p2.validate(ma))
      p2.first_name = "Winston"
      self.assertTrue(p2.validate(ma))

      ma.rollback()

  def test_numeric(self):
    v = Validation.is_numeric("age")
    self.assertTrue(v.is_valid(self.model)[0])

    v = Validation.is_numeric("age", numtype="int")
    self.model.age = "68.5"
    self.assertFalse(v.is_valid(self.model)[0])

    v = Validation.is_numeric("age", numtype="int", min=0, max=120)
    self.model.age = -1
    self.assertFalse(v.is_valid(self.model)[0])
    self.model.age = 121
    self.assertFalse(v.is_valid(self.model)[0])
    self.model.age = 68
    self.assertTrue(v.is_valid(self.model)[0])

    self.model.age = None
    self.assertFalse(v.is_valid(self.model)[0])

  def test_date(self):
    DTT = namedtuple("dtt", ["date", "datetime"])
    dtt = DTT("2014-01-01", "2014-01-01 13:45:15.0")
    vd = Validation.is_date("date")
    vdt = Validation.is_datetime("datetime")

    self.assertTrue(vd.is_valid(dtt)[0])
    self.assertTrue(vdt.is_valid(dtt)[0])

  def test_is_in(self):
    v = Validation.is_in("first_name", { "john", "paul", "george", "ringo" })
    self.assertFalse(v.is_valid(self.model)[0])
    self.model.first_name = "ringo"
    self.assertTrue(v.is_valid(self.model)[0])

  def test_filter(self):
    class M(ModelBase):
      attrs = dict(first_name=None, last_name=None)

    m1 = M(first_name="trey", last_name=None)
    m2 = M(first_name=None, last_name=None)
    v = Validation.is_present("last_name").add_filter(lambda m: m.first_name is not None)

    self.assertFalse(v.validate(m1))
    self.assertTrue(v.validate(m2))



import re
import unittest

from bidon.data import ModelBase, Validation, Validator


class Person(ModelBase):
  table_name = "people"
  attrs = {
    "first_name": None,
    "last_name": None,
    "age": 0 }
  validator = Validator([
    Validation.is_present("last_name"),
    Validation.is_present("first_name")])


class DataModelBaseTestCase(unittest.TestCase):
  def test_init(self):
    p1 = Person(dict(first_name="Lucy", last_name="In the Sky With Diamonds"))
    self.assertTrue(hasattr(p1, "first_name"))
    self.assertTrue(hasattr(p1, "last_name"))
    self.assertTrue(hasattr(p1, "age"))
    p2 = Person()
    self.assertTrue(hasattr(p2, "first_name"))
    self.assertTrue(hasattr(p2, "last_name"))
    self.assertTrue(hasattr(p2, "age"))
    self.assertEqual(p2.age, 0)
    p3 = Person(id=11)
    self.assertEqual(p3.id, 11)

  def test_is_new(self):
    class C1(ModelBase):
      primary_key_name = "id"

    class C2(ModelBase):
      primary_key_name = ("parent_id", "name")

    c1 = C1()
    c2 = C2()
    self.assertTrue(c1.is_new)
    self.assertTrue(c2.is_new)
    c1.update(id=1)
    c2.update(parent_id=1, name="Hi")
    self.assertFalse(c1.is_new)
    self.assertFalse(c2.is_new)

  def test_validate(self):
    p1 = Person(dict(first_name="Eleanor", last_name="Rigby"))
    self.assertTrue(p1.validate())
    p1.first_name = None
    self.assertFalse(p1.validate())
    p1.first_name = "Eleanor"
    p1.last_name = None
    self.assertFalse(p1.validate())

  def test_error_methods(self):
    p1 = Person(dict(first_name="Mother Nature's", last_name="Son"))
    self.assertFalse(p1.has_errors)
    p1.add_error("first_name", "Isn't real")
    self.assertTrue(p1.has_errors)
    p1.clear_errors()
    self.assertFalse(p1.has_errors)

  def test_update(self):
    p1 = Person(dict(first_name="Rocky", last_name="Raccoon"))
    self.assertEqual(1, len(p1.update(dict(last_name="Balboa"))))
    self.assertEqual(p1.last_name, "Balboa")
    self.assertEqual(0, len(p1.update(dict(first_name="Rocky"))))
    self.assertEqual({"first_name", "last_name", "age"}, p1.update(first_name="A", last_name="B", age=44))
    ModelBase.strict_attrs = True
    with self.assertRaises(Exception):
      p1.update(some_unknown_attr="hihi")

  def test_to_dict(self):
    p1 = Person(dict(first_name="Mr.", last_name="Kite"))
    self.assertEqual(dict(first_name="Mr.", last_name="Kite", age=0, errors={}, updated_at=None, created_at=None, id=None), p1.to_dict())
    self.assertEqual(dict(first_name="Mr."), p1.to_dict(include_keys=["first_name"]))
    self.assertEqual(dict(first_name="Mr.", last_name="Kite", age=0), p1.to_dict(exclude_keys=["errors", "updated_at", "created_at", "id"]))




import unittest
from xml.etree import ElementTree as ET

import bidon.util.transform as T
import bidon.util.convert as conv
from bidon.data import ForeignModelWrapper


class Person(ForeignModelWrapper):
  attrs = dict(first_name=None, last_name=None, company_id=None)
  transform_args = (
    None,
    dict(first_name=T.val(T.json_val("/first_name")),
         last_name=T.val(T.json_val("/last_name")),
         company_id=T.val(T.json_val("/company/id"), conv.to_int)))


class SimplePerson(ForeignModelWrapper):
  attrs = dict(id=None, name=None)
  transform_args = (
    None,
    dict(id=T.val(T.json_val("/id"), conv.to_int),
         name=T.val(T.json_val("/name"))))


class CompanyXML(ForeignModelWrapper):
  attrs = dict(id=None, name=None)


class PersonXML(ForeignModelWrapper):
  attrs = dict(id=None, first_name=None, last_name=None, company=None)
  transform_args = (
    None,
    dict(id=T.val(T.xml_attr("id"), conv.to_int),
         first_name=T.val(T.xml_text("first_name")),
         last_name=T.val(T.xml_text("last_name")),
         company=T.obj(T.xml_child("company"),
                       dict(id=T.val(T.xml_attr("id"), conv.to_int),
                            name=T.val(T.xml_text("name"))),
                       CompanyXML)))


class DataForeignModelWrapperTestCase(unittest.TestCase):
  def test_create_raw(self):
    d = {
      "first_name": "Trey",
      "last_name": "Cucco",
      "company": {
        "id": 188,
        "name": "ACME" }}
    p = Person.create(d)
    self.assertEqual(p.first_name, "Trey")
    self.assertEqual(p.last_name, "Cucco")
    self.assertEqual(p.company_id, 188)

  def test_map(self):
    l = [dict(id=1, name="Trey Cucco"),
         dict(id=2, name="J.R.R. Tolkien"),
         dict(id=3, name="C.S. Lewis")]
    ml = list(SimplePerson.map(l))

    self.assertEqual(len(ml), 3)
    for s in ml:
      self.assertIsInstance(s, SimplePerson)
    self.assertEqual(ml[0].id, 1)
    self.assertEqual(ml[0].name, "Trey Cucco")
    self.assertEqual(ml[1].id, 2)
    self.assertEqual(ml[1].name, "J.R.R. Tolkien")
    self.assertEqual(ml[2].id, 3)
    self.assertEqual(ml[2].name, "C.S. Lewis")

  def test_create_xml(self):
    xmlt = """<person id="1">
      <first_name>Trey</first_name>
      <last_name>Cucco</last_name>
      <company id="1">
        <name>ACME</name>
      </company>
    </person>"""
    root = ET.fromstring(xmlt)
    p = PersonXML.create(root)
    self.assertEqual(p.id, 1)
    self.assertIsInstance(p.company, CompanyXML)
    self.assertEqual(p.company.id, 1)
    self.assertEqual(p.company.name, "ACME")
    self.assertEqual(p.first_name, "Trey")
    self.assertEqual(p.last_name, "Cucco")
